import contextlib
from contextlib import contextmanager
from functools import wraps
import hashlib
import io
import logging
import os
from pathlib import Path
import shutil
import tempfile
from typing import Any, Generator
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen
from uuid import uuid4

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from django.conf import settings
from django.contrib.gis.db.models import Model
from django.core.files import File
from django.db.models.fields.files import FieldFile
from django.utils.safestring import mark_safe
from filelock import FileLock, Timeout
import psutil

try:
    from minio_storage.storage import MinioStorage
except ImportError:
    MinioStorage = None

logger = logging.getLogger(__name__)


def get_temp_dir():
    path = Path(getattr(settings, 'RGD_TEMP_DIR', os.path.join(tempfile.gettempdir(), 'rgd')))
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_cache_dir():
    path = Path(get_temp_dir(), 'file_cache')
    path.mkdir(parents=True, exist_ok=True)
    return path


@contextmanager
def safe_urlopen(url: str, *args, **kwargs):
    with contextlib.closing(urlopen(url, *args, **kwargs)) as remote:
        yield remote


def compute_hash(handle: io.BufferedIOBase, chunk_num_blocks: int = 128):
    sha = hashlib.sha512()
    while chunk := handle.read(chunk_num_blocks * sha.block_size):
        sha.update(chunk)
    return sha.hexdigest()


def compute_checksum_file_field(field_file: FieldFile, chunk_num_blocks: int = 128):
    with field_file.open() as f:
        return compute_hash(f, chunk_num_blocks)


def compute_checksum_url(url: str, chunk_num_blocks: int = 128):
    with safe_urlopen(url) as remote:
        return compute_hash(remote, chunk_num_blocks)


def _link_url(obj: Model, field: str):
    if not getattr(obj, field, None):
        return 'No attachment'
    attr = getattr(obj, field)
    if callable(attr):
        url = attr()
    else:
        url = attr.url
    return mark_safe(f'<a href="{url}" download>Download</a>')


def get_or_create_no_commit(model: Model, defaults: dict = None, **kwargs):
    try:
        return model.objects.get(**kwargs), False
    except model.DoesNotExist:
        if not defaults:
            defaults = {}
        defaults.update(kwargs)
        return model(**defaults), True


def get_s3_client():
    if boto3.session.Session().get_credentials():
        s3 = boto3.client('s3')
    else:
        # No credentials present (often the case in dev), use unsigned requests
        s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    return s3


def _download_url_file_to_stream(
    url: str, dest_stream: io.BufferedIOBase, num_blocks: int = 128, block_size: int = 128
):
    parsed = urlparse(url)
    if parsed.scheme == 's3':
        s3 = get_s3_client()
        s3.download_fileobj(parsed.netloc, parsed.path.lstrip('/'), dest_stream)
    else:
        with safe_urlopen(url) as remote:
            while chunk := remote.read(num_blocks * block_size):
                dest_stream.write(chunk)
                dest_stream.flush()


def download_url_file_to_local_path(
    url: str,
    path: str,
    num_blocks: int = 128,
    block_size: int = 128,
) -> Path:
    dest_path = Path(path)
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    with open(dest_path, 'wb') as dest_stream:
        _download_url_file_to_stream(url, dest_stream, num_blocks=num_blocks, block_size=block_size)
    return Path(dest_path)


def precheck_fuse(url: str) -> bool:
    try:
        import simple_httpfs  # noqa
    except (ImportError, EnvironmentError):
        return False
    parsed = urlparse(url)
    if parsed.scheme not in ['https', 'http']:
        return False
    try:
        # The FUSE lib will not catch URL errors
        with safe_urlopen(url) as _:
            pass
    except HTTPError:
        return False
    return True


@contextmanager
def url_file_to_fuse_path(url: str) -> Generator[Path, None, None]:
    # Could raise ValueError within context
    # Assumes `precheck_fuse` was verified prior
    # See https://github.com/ResonantGeoData/ResonantGeoData/issues/237
    parsed = urlparse(url)
    if parsed.scheme == 'https':
        fuse_path = url.replace('https://', '/tmp/rgd/https/') + '..'
    elif parsed.scheme == 'http':
        fuse_path = url.replace('http://', '/tmp/rgd/http/') + '..'
    else:
        raise ValueError(f'Scheme {parsed.scheme} not currently handled by FUSE.')
    logger.info(f'FUSE path: {fuse_path}')
    yield Path(fuse_path)


@contextmanager
def patch_internal_presign(f: FieldFile):
    """Create an environment where Minio-based `FieldFile`s construct a locally accessible presigned URL.

    Sometimes the external host differs from the internal host for Minio files (e.g. in development).
    Getting the URL in this context ensures that the presigned URL returns the correct host for the
    odd situation of accessing the file locally.
    """
    if (
        MinioStorage is not None
        and isinstance(f.storage, MinioStorage)
        and getattr(settings, 'MINIO_STORAGE_MEDIA_URL', None) is not None
    ):
        original_base_url = f.storage.base_url
        try:
            f.storage.base_url = None
            yield
        finally:
            f.storage.base_url = original_base_url
        return
    yield


@contextmanager
def output_path_helper(filename: str, output: FieldFile):
    workdir = get_temp_dir()
    with tempfile.TemporaryDirectory(dir=workdir) as tmpdir:
        output_path = os.path.join(tmpdir, filename)
        try:
            # Yield the path for the user to perform a task
            yield output_path
        except Exception as e:
            raise e
        else:
            # Save the file contents to the output field only on success
            with open(output_path, 'rb') as f:
                output.save(os.path.basename(output_path), f)


@contextmanager
def input_output_path_helper(source, output: FieldFile, prefix: str = '', suffix: str = ''):
    """Yield source and output paths between a ChecksumFile and a FileFeild.

    The output path is saved to the output field after yielding.

    """
    filename = prefix + os.path.basename(source.name) + suffix
    with source.yield_local_path() as file_path:
        filename = prefix + os.path.basename(source.name) + suffix
        with output_path_helper(filename, output) as output_path:
            try:
                # Yield the paths for the user to perform a task
                yield (file_path, output_path)
            except Exception as e:
                raise e
            else:
                # Save the file contents to the output field only on success
                with open(output_path, 'rb') as f:
                    output.save(os.path.basename(output_path), f)


def uuid_prefix_filename(instance: Any, filename: str):
    """Use a variable in settings to add a prefix to the path and keep the random uuid."""
    prefix = getattr(settings, 'RGD_FILE_FIELD_PREFIX', None)
    if prefix:
        return f'{prefix}/{uuid4()}/{filename}'
    return f'{uuid4()}/{filename}'


def skip_signal():
    """Skip the signal on an instance-basis."""

    def _skip_signal(signal_func):
        @wraps(signal_func)
        def _decorator(sender, instance, *args, **kwargs):
            if hasattr(instance, 'skip_signal') and instance.skip_signal:
                return None
            return signal_func(sender, instance, *args, **kwargs)

        return _decorator

    return _skip_signal


def download_field_file_to_local_path(field_file: FieldFile, path: str) -> Path:
    """Download entire FieldFile to disk location.

    This overrides `girder_utils.field_file_to_local_path` to download file to
    local path without a context manager. Cleanup must be handled by caller.

    """
    dest_path = Path(path)
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    the_path = Path(dest_path)
    with field_file.open('rb'):
        file_obj: File = field_file.file
        if type(file_obj) is File:
            # When file_obj is an actual File, (typically backed by FileSystemStorage),
            # it is already at a stable path on disk.
            # We must symlink it into the desired path
            os.symlink(file_obj.name, dest_path)
            return the_path
        else:
            # When file_obj is actually a subclass of File, it only provides a Python
            # file-like object API. So, it must be copied to a stable path.
            with open(dest_path, 'wb') as dest_stream:
                shutil.copyfileobj(file_obj, dest_stream)
                dest_stream.flush()
            return the_path


def clean_file_cache(override_target=None):
    """Clean the file cache to achieve RGD_TARGET_AVAILABLE_CACHE (Gb).

    Return
    ------
    A tuple of the starting and ending free space in bytes.

    """
    cache = get_cache_dir()
    # Sort each directory by mtime
    paths = sorted(
        [f for f in Path(cache).iterdir() if not str(f).endswith('.lock')], key=os.path.getmtime
    )  # This sorts oldest to latest
    # While free space is not enough, remove directories until all ar gone
    initial = psutil.disk_usage(cache).free
    target = override_target or getattr(settings, 'RGD_TARGET_AVAILABLE_CACHE', 2)
    logger.info(f'Cleaning file cache... Starting free space is {initial} bytes.')
    while psutil.disk_usage(cache).free * 1e-9 < target:
        if not len(paths):
            # If we delete everything and still cannot acheive target, warn
            logger.error(
                f'Target cache free space of {target * 1e9} bytes not achieved when empty.'
            )
            break
        path = paths.pop(0)
        # Check if resource is locked and skip if so
        lock_path = Path(f'{path}.lock')
        lock = FileLock(lock_path)
        try:
            with lock.acquire(timeout=0):
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)
            # Remove the lockfile as well
            os.remove(lock_path)
            logger.info(f'removed: {path}')
        except Timeout:
            # Another task holds the lock on this file/directory: do not delete
            logger.info(f'File is locked, skipping: {path}')
            pass
    free = psutil.disk_usage(cache).free
    logger.debug(f'Finished cleaning file cache. Available free space is {free} bytes.')
    return initial, free


def purge_file_cache():
    """Completely purge all files from the file cache.

    This should be used with extreme caution, it will delete files that could
    be in use.

    """
    cache = get_cache_dir()
    shutil.rmtree(cache)
    cache = get_cache_dir()  # Return the cache dir so that a fresh directory is created.
    logger.debug(
        f'Purged file cache. Available free space is {psutil.disk_usage(cache).free} bytes.'
    )
    return cache



from django.http import HttpResponseRedirect
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import action
from rgd import models, serializers
from rgd.filters import SpatialEntryFilter
from rgd.rest.base import ModelViewSet, ReadOnlyModelViewSet


class CollectionViewSet(ModelViewSet):
    serializer_class = serializers.CollectionSerializer
    queryset = models.Collection.objects.all()


class CollectionPermissionViewSet(ModelViewSet):
    serializer_class = serializers.CollectionPermissionSerializer
    queryset = models.CollectionPermission.objects.all()


class ChecksumFileViewSet(ModelViewSet):
    serializer_class = serializers.ChecksumFileSerializer
    queryset = models.ChecksumFile.objects.all()

    @swagger_auto_schema(
        method='GET',
        operation_summary='Download ChecksumFile data directly from S3.',
    )
    @action(detail=True)
    def data(self, request, pk=None):
        obj = self.get_object()
        return HttpResponseRedirect(obj.get_url())


class SpatialEntryViewSet(ReadOnlyModelViewSet):
    serializer_class = serializers.SpatialEntrySerializer
    queryset = models.SpatialEntry.objects.all()
    filterset_class = SpatialEntryFilter

    @swagger_auto_schema(
        method='GET',
        operation_summary='Get the footprint of this SpatialEntry.',
    )
    @action(detail=True, serializer_class=serializers.SpatialEntryFootprintSerializer)
    def footprint(self, request, pk=None):
        return self.retrieve(request, pk=pk)


class SpatialAssetViewSet(ModelViewSet):
    serializer_class = serializers.SpatialAssetSerializer
    queryset = models.SpatialAsset.objects.all()



from typing import List

from django.apps import apps
from django.conf import settings
from django.contrib.auth.backends import BaseBackend
from django.core.exceptions import PermissionDenied
from django.db.models import Model, Q
from rest_framework import filters, permissions
from rgd import models


# get django model via name, regardless of app
def get_model(model_name: str):
    for m in apps.get_models():
        if m.__name__ == model_name:
            return m


def get_subclasses(model: Model):
    """Retrieve all model subclasses for the provided class excluding the model class itself."""
    return set(
        [
            m
            for m in apps.get_models()
            if issubclass(m, model)
            and m != model
            and issubclass(m, models.mixins.PermissionPathMixin)
        ]
    )


def get_permissions_paths(model: Model, target_model: Model) -> List[str]:
    """Get all possible paths to the 'target_model'.

    Produces relationships represented as 'dunder's ('__').
    """
    if model == target_model:
        return ['']

    if model == models.SpatialEntry:
        submodels = get_subclasses(model)
        paths = []
        for sub in submodels:
            # TODO: there should be a cleaner way to do this.
            model_name = sub.__name__.lower()
            [paths.append(f'{model_name}__{s}') for s in get_permissions_paths(sub, target_model)]
        return paths

    if not issubclass(model, models.mixins.PermissionPathMixin):
        raise TypeError(f'{type(model)} does not have the `PermissionPathMixin` interface.')

    paths = []
    for path in model.permissions_paths:
        field, next_model = path

        if type(next_model) == str:

            # grab model via class name
            next_model = get_model(next_model)

        # sanity check
        if not issubclass(next_model, Model):
            raise TypeError('Failed to extract next_model in permissions_paths')

        if next_model == target_model:
            paths.append(field)
        else:
            next_path = get_permissions_paths(next_model, target_model)
            for n in next_path:
                paths.append(f'{field}__{n}')

    return paths


def filter_perm(user, queryset, role):
    """Filter a queryset."""
    if queryset.model == models.SpatialEntry:
        # Return subclasses of SpatialEntry
        queryset = queryset.select_subclasses()
    # Called outside of view
    if user is None:
        return queryset
    # Must be logged in
    if not user.is_active or user.is_anonymous:
        return queryset.none()
    # Superusers can see all (not staff users)
    if user.is_active and user.is_superuser:
        return queryset
    # Check permissions
    subquery = queryset.none()
    # Now grab resources created by this user - only file-based models
    if not issubclass(queryset.model, (models.CollectionPermission, models.Collection)):
        for path in get_permissions_paths(queryset.model, models.ChecksumFile):
            created_by_path = (path + '__' if path != '' else path) + 'created_by'
            condition = Q(**{created_by_path: user})
            subquery = subquery.union(queryset.filter(condition).values('pk'))
    for path in get_permissions_paths(queryset.model, models.CollectionPermission):
        # `path` can be an empty string (meaning queryset is `CollectionPermission`)
        user_path = (path + '__' if path != '' else path) + 'user'
        role_path = (path + '__' if path != '' else path) + 'role'
        condition = Q(**{user_path: user}) & Q(**{role_path + '__gte': role})
        if getattr(settings, 'RGD_GLOBAL_READ_ACCESS', False):
            condition |= Q(**{path + '__isnull': True})
        subquery = subquery.union(queryset.filter(condition).values('pk'))
    return queryset.filter(pk__in=subquery)


def filter_read_perm(user, queryset):
    """Filter a queryset to what the user may read."""
    return filter_perm(user, queryset, models.CollectionPermission.READER)


def filter_write_perm(user, queryset):
    """Filter a queryset to what the user may edit."""
    return filter_perm(user, queryset, models.CollectionPermission.OWNER)


def check_read_perm(user, obj):
    """Raise 'PermissionDenied' error if user does not have read permissions."""
    model = type(obj)
    if not filter_read_perm(user, model.objects.filter(pk=obj.pk)).exists():
        raise PermissionDenied


def check_write_perm(user, obj):
    """Raise 'PermissionDenied' error if user does not have write permissions."""
    # Called outside of view
    model = type(obj)
    if not filter_write_perm(user, model.objects.filter(pk=obj.pk)).exists():
        raise PermissionDenied


class CollectionAuthorizationBackend(BaseBackend):
    def has_perm(self, user, perm, obj=None):
        """Supplement default Django permission backend.

        Returns `True` if the user has the specified permission, where perm is in the format
        `"<app label>.<permission codename>"`. If the user is
        inactive, this method will always return False. For an active superuser, this method
        will always return `True`.

        https://docs.djangoproject.com/en/3.1/ref/contrib/auth/#django.contrib.auth.models.User.has_perm
        """
        app_label, codename = perm.split('.')
        if app_label == 'geodata':
            if codename.startswith('view'):
                check_read_perm(user, obj)
            if (
                codename.startswith('add')
                or codename.startswith('delete')
                or codename.startswith('change')
            ):
                check_write_perm(user, obj)


class CollectionAuthorization(permissions.BasePermission):
    """A custom Django REST Framework permission backend.

    Object-level permission to only allow owners of an object to edit it
    and readers to view it. This will not filter the queryset in 'list' views.
    You should use this in combination with the `CollectionAuthorizationFilter` to
    also filter the queryset in 'list' views.

    https://www.django-rest-framework.org/api-guide/permissions/#custom-permissions
    """

    def has_object_permission(self, request, view, obj):
        # Check read permissions for GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            check_read_perm(request.user, obj)
        else:
            check_write_perm(request.user, obj)
        # The above calls should raise an exception if the user does not have permissions.
        return True


class CollectionAuthorizationFilter(filters.BaseFilterBackend):
    """A custom Django REST Framework filter backend for permissions.

    As a result of applying this filter, objects that the user does not have persmissions
    to view will return a '404' since the queryset is filtered before object-level
    permissions are checked.

    https://www.django-rest-framework.org/api-guide/permissions/#custom-permissions
    """

    def filter_queryset(self, request, queryset, view):
        # Check read permissions for GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return filter_read_perm(request.user, queryset)
        # Otherwise, assume this is a writable query.
        return filter_write_perm(request.user, queryset)



import json

import pytest
from rgd.models import ChecksumFile
from rgd_client import RgdClient


@pytest.mark.django_db(transaction=True)
def test_basic_search(py_client: RgdClient, spatial_asset_a):
    bbox = {
        'type': 'Polygon',
        'coordinates': [
            [
                [-84.10240173339842, 39.72989765591686],
                [-84.0395736694336, 39.72989765591686],
                [-84.0395736694336, 39.78057429170315],
                [-84.10240173339842, 39.78057429170315],
                [-84.10240173339842, 39.72989765591686],
            ]
        ],
    }

    q = py_client.rgd.search(query=json.dumps(bbox), predicate='intersects')
    assert len(q) == 1


@pytest.mark.django_db(transaction=True)
def test_create_file_from_url(py_client: RgdClient, checksum_file_url: ChecksumFile):
    """Test that creation of a ChecksumFile with a URL succeeds."""
    file_dict = py_client.rgd.create_file_from_url(checksum_file_url.get_url())
    assert file_dict['url'] == checksum_file_url.get_url()
    assert file_dict['type'] == 2



from django.shortcuts import get_object_or_404  # , render
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rgd.permissions import check_read_perm, get_model


def _get_status_response(request, model, pk):
    model_class = get_model(model)
    if not model_class:
        raise AttributeError(f'Model ({model}) does not exist.')
    instance = get_object_or_404(model_class, pk=pk)
    check_read_perm(request.user, instance)
    if not hasattr(instance, 'status'):
        raise AttributeError(f'Model ({model}) has no attribute (status).')
    data = {
        'pk': instance.pk,
        'model': model,
        'status': instance.status,
    }
    return Response(data)


@swagger_auto_schema(
    method='GET',
    operation_summary='Check the status.',
)
@api_view(['GET'])
def get_status(request, model, pk):
    """Get the status of any TaskEventMixin model."""
    return _get_status_response(request, model, pk)
